@@ -15,15 +15,20 @@
 """
 from __future__ import print_function
 
-import subprocess
 import errno
 import gc
+import io
+import os
 import re
+import subprocess
 import sys
 import threading
 import time
 import traceback
 import unittest
+import warnings
+
+from contextlib import contextmanager
 
 from six import StringIO
 from zope.testrunner.find import import_name
@@ -59,7 +64,7 @@ class UnexpectedSuccess(Exception):
     pass
 
 
-PYREFCOUNT_PATTERN = re.compile('\[[0-9]+ refs\]')
+PYREFCOUNT_PATTERN = re.compile(r'\[[0-9]+ refs\]')
 
 is_jython = sys.platform.startswith('java')
 
@@ -87,10 +92,19 @@ class Runner(object):
     executing tests as well as configuring itself from the (command-line)
     options passed into it.
 
+    .. versionchanged:: 4.8.0
+       Add the *warnings* keyword argument. If this is ``None`` (the default)
+       and the user hasn't configured Python otherwise with command-line arguments
+       or environment variables, we will enable the default warnings, including
+       ``DeprecationWarning``, when running tests. Otherwise, it can be any
+       string acceptable to :func:`warnings.simplefilter` and that filter will
+       be in effect while running tests.
+
     """
 
     def __init__(self, defaults=None, args=None, found_suites=None,
-                 options=None, script_parts=None, cwd=None):
+                 options=None, script_parts=None, cwd=None,
+                 warnings=None):
         if defaults is None:
             self.defaults = []
         else:
@@ -101,6 +115,18 @@ class Runner(object):
         self.script_parts = script_parts
         self.cwd = cwd
         self.failed = True
+        if warnings is None and not sys.warnoptions:
+            # even if DeprecationWarnings are ignored by default
+            # print them anyway unless other warnings settings are
+            # specified by the warnings arg or the -W python flag
+            self.warnings = 'default'
+        else:
+            # here self.warnings is set either to the value passed
+            # to the warnings args or to None.
+            # If the user didn't pass a value self.warnings will
+            # be None. This means that the behavior is unchanged
+            # and depends on the values passed to -W.
+            self.warnings = warnings
 
         self.ran = 0
         self.skipped = []
@@ -143,32 +169,37 @@ class Runner(object):
         self.layer_name_cache = _layer_name_cache
         self.layer_name_cache.clear()
 
-        # Global setup
-        for feature in self.features:
-            feature.global_setup()
+        with self._enabled_warnings():
+            # Enable warnings during setup so that
+            # warnings raised on import (which we do for test
+            # discover) can be reported.
 
-        # Late setup
-        #
-        # Some system tools like profilers are really bad with stack frames.
-        # E.g. hotshot doesn't like it when we leave the stack frame that we
-        # called start() from.
-        for feature in self.features:
-            feature.late_setup()
-
-        try:
-            if self.do_run_tests:
-                self.run_tests()
-        finally:
-            # Early teardown
-            for feature in reversed(self.features):
-                feature.early_teardown()
-            # Global teardown
-            for feature in reversed(self.features):
-                feature.global_teardown()
-
-        if self.show_report:
+            # Global setup
             for feature in self.features:
-                feature.report()
+                feature.global_setup()
+
+            # Late setup
+            #
+            # Some system tools like profilers are really bad with stack frames.
+            # E.g. hotshot doesn't like it when we leave the stack frame that we
+            # called start() from.
+            for feature in self.features:
+                feature.late_setup()
+
+            try:
+                if self.do_run_tests:
+                    self.run_tests()
+            finally:
+                # Early teardown
+                for feature in reversed(self.features):
+                    feature.early_teardown()
+                # Global teardown
+                for feature in reversed(self.features):
+                    feature.global_teardown()
+
+            if self.show_report:
+                for feature in self.features:
+                    feature.report()
 
     def configure(self):
         if self.args is None:
@@ -222,6 +253,26 @@ class Runner(object):
         # Remove all features that aren't activated
         self.features = [f for f in self.features if f.active]
 
+    @contextmanager
+    def _enabled_warnings(self):
+        """
+        A context manager to enable warnings as configured.
+        """
+        with warnings.catch_warnings():
+            if self.warnings:
+                # if self.warnings is set, use it to filter all the warnings
+                warnings.simplefilter(self.warnings)
+                # if the filter is 'default' or 'always', special-case the
+                # warnings from the deprecated unittest methods to show them
+                # no more than once per module, because they can be fairly
+                # noisy.  The -Wd and -Wa flags can be used to bypass this
+                # only when self.warnings is None.
+                if self.warnings in ['default', 'always']:
+                    warnings.filterwarnings('module',
+                                            category=DeprecationWarning,
+                                            message=r'Please use assert\w+ instead.')
+            yield
+
     def run_tests(self):
         """Run all tests that were registered.
 
@@ -429,7 +480,7 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
                               layer_name, layer, failures, errors, skipped,
                               resume_number, cwd=None):
     output = options.output
-
+    child = None
     try:
         # BBB
         if script_parts is None:
@@ -533,6 +584,33 @@ def spawn_layer_in_subprocess(result, script_parts, options, features,
 
     finally:
         result.done = True
+        if child is not None:
+            # Regardless of whether the process ran to completion, we
+            # must properly cleanup the process to avoid
+            # `ResourceWarning: subprocess XXX is still alive` and
+            # `ResourceWarning: unclosed file` for its stdout and
+            # stderr.
+            child.kill()
+            child.communicate()
+
+
+def _get_output_buffer(stream):
+    """Get a binary-safe version of a stream."""
+    try:
+        fileno = stream.fileno()
+    except (io.UnsupportedOperation, AttributeError):
+        pass
+    else:
+        # Win32 mangles \r\n to \n and that breaks streams.  See
+        # https://bugs.launchpad.net/bugs/505078.
+        if sys.platform == 'win32':
+            import msvcrt
+            msvcrt.setmode(fileno, os.O_BINARY)
+    try:
+        stream.write(b'')
+    except TypeError:
+        return stream.buffer
+    return stream
 
 
 class AbstractSubprocessResult(object):
@@ -554,8 +632,6 @@ class DeferredSubprocessResult(AbstractSubprocessResult):
     """Keeps stdout around for later processing,"""
 
     def write(self, out):
-        if not isinstance(out, str): # It's binary, which means it's Python 3
-            out = out.decode()
         if not _is_dots(out):
             self.stdout.append(out)
 
@@ -563,18 +639,17 @@ class DeferredSubprocessResult(AbstractSubprocessResult):
 class ImmediateSubprocessResult(AbstractSubprocessResult):
     """Sends complete output to queue."""
 
+    def __init__(self, layer_name, queue):
+        super(ImmediateSubprocessResult, self).__init__(layer_name, queue)
+        self.stream = _get_output_buffer(sys.stdout)
+
     def write(self, out):
-        if not isinstance(out, str):
-            # In Python 3, a Popen process stdout uses bytes,
-            # While normal stdout uses strings. So we need to convert
-            # from bytes to strings here.
-            out = out.decode()
-        sys.stdout.write(out)
+        self.stream.write(out)
         # Help keep-alive monitors (human or automated) keep up-to-date.
-        sys.stdout.flush()
+        self.stream.flush()
 
 
-_is_dots = re.compile(r'\.+(\r\n?|\n)').match # Windows sneaks in a \r\n.
+_is_dots = re.compile(br'\.+(\r\n?|\n)').match # Windows sneaks in a \r\n.
 class KeepaliveSubprocessResult(AbstractSubprocessResult):
     "Keeps stdout for later processing; sends marks to queue to show activity."
 
@@ -587,8 +662,6 @@ class KeepaliveSubprocessResult(AbstractSubprocessResult):
     done = property(lambda self: self._done, _set_done)
 
     def write(self, out):
-        if not isinstance(out, str): # It's binary, which means it's Python 3
-            out = out.decode()
         if _is_dots(out):
             self.queue.put((self.layer_name, out.strip()))
         else:
@@ -601,7 +674,8 @@ def resume_tests(script_parts, options, features, layers, failures, errors,
     stdout_queue = None
     if options.processes == 1:
         result_factory = ImmediateSubprocessResult
-    elif options.verbose > 1:
+    elif (options.verbose > 1 and
+            not options.subunit and not options.subunit_v2):
         result_factory = KeepaliveSubprocessResult
         stdout_queue = Queue.Queue()
     else:
@@ -623,6 +697,7 @@ def resume_tests(script_parts, options, features, layers, failures, errors,
     current_result = next(results_iter)
     last_layer_intermediate_output = None
     output = None
+    stdout = _get_output_buffer(sys.stdout)
     while ready_threads or running_threads:
         while len(running_threads) < options.processes and ready_threads:
             thread = ready_threads.pop(0)
@@ -643,18 +718,19 @@ def resume_tests(script_parts, options, features, layers, failures, errors,
             if layer_name != last_layer_intermediate_output:
                 # Clarify what layer is reporting activity.
                 if previous_output is not None:
-                    sys.stdout.write(']\n')
-                sys.stdout.write(
-                    '[Parallel tests running in %s:\n  ' % (layer_name,))
+                    stdout.write(b']\n')
+                stdout.write(
+                    ('[Parallel tests running in '
+                     '%s:\n  ' % (layer_name,)).encode('utf-8'))
                 last_layer_intermediate_output = layer_name
-            sys.stdout.write(output)
+            stdout.write(output)
         # Display results in the order they would have been displayed, had the
         # work not been done in parallel.
         while current_result and current_result.done:
             if output is not None:
-                sys.stdout.write(']\n')
+                stdout.write(b']\n')
                 output = None
-            sys.stdout.writelines(current_result.stdout)
+            stdout.writelines(current_result.stdout)
 
             try:
                 current_result = next(results_iter)
@@ -662,7 +738,7 @@ def resume_tests(script_parts, options, features, layers, failures, errors,
                 current_result = None
 
         # Help keep-alive monitors (human or automated) keep up-to-date.
-        sys.stdout.flush()
+        stdout.flush()
         time.sleep(0.01) # Keep the loop from being too tight.
 
     # Return the total number of tests run.
@@ -900,7 +976,7 @@ def layer_sort_key(layer):
         # We make the simplifying assumption that the order of initialization
         # of base layers does not matter.  Given that, traversing the bases
         # in reverse order here keeps the ordering of layers in
-        # testrunner-layers.txt the same as it was in older versions of
+        # testrunner-layers.rst the same as it was in older versions of
         # zope.testrunner, so let's use that.
         for base in layer.__bases__[::-1]:
             if base is not object and base not in seen:
